################################################################################
##                                                                            ##
##      Title: Install-HSA.ps1                                                ##
##  Publisher: Lenovo                                                         ##
##    Version: 1.02                                                           ##
##       Date: 2020-06-03                                                     ##
##                                                                            ##
##                              Legal Disclaimer                              ##
##                                                                            ##
##  The sample scripts are not supported under any Lenovo standard support    ##
##  program or service. The sample scripts are provided AS IS without         ##
##  warranty of any kind. Lenovo further disclaims all implied warranties     ##
##  including, without limitation, any implied warranties of                  ##
##  merchantability or of fitness for a particular purpose. The entire risk   ##
##  arising out of the use or performance of the sample scripts and           ##
##  documentation remains with you. In no event shall Lenovo, its authors,    ##
##  or anyone else involved in the creation, production, or delivery of the   ##
##  scripts be liable for any damages whatsoever (including, without          ##
##  limitation, damages for loss of business profits, business                ##
##  interruption, loss of business information, or other pecuniary loss)      ##
##  arising out of the use of or inability to use the sample scripts or       ##
##  documentation, even if Lenovo has been advised of the possibility of      ##
##  such damages.                                                             ##
##                                                                            ##
################################################################################

# https://thinkdeploy.blogspot.com/2020/06/hardware-support-apps-without-microsoft.html

<#
.SYNOPSIS

    List the names of or install Hardware Support Applications (HSA) by 
    reading manifest files located in the subdirectories.

.DESCRIPTION

    By reading manifest files found in the subdirectories of a Hardware Support 
    Applications Pack, this script allows for the deployment of one, many, or 
    all Hardware Support Applications for a given model.  All installations are 
    against an offline installation of Windows 10.

.PARAMETER LIST

    Used to list the names of all Hardware Support Applications in the 
    subdirectories.  Returns the list of the Hardware Support 
    Applications names to the screen.

    -LIST

.PARAMETER EXPORT

    Used to list the names of all Hardware Support Applications in the 
    subdirectories.  Returns the list of names to the Export_<Date><Time>.log 
    file in the directory from which the script was executed.

    -LIST -EXPORT

.PARAMETER OFFLINE

    Used to install the Hardware Support Applications.

    -OFFLINE <-ALL, FILE '<FileName>.txt', or -NAME '<HSA Name>'>

.PARAMETER ALL

    Used to install all Hardware Support Applications in the subdirectories 
    below the script.

    -OFFLINE -ALL

.PARAMETER FILE

    Used to install a list of Hardware Support Applications.  The text file 
    should be formatted with one Hardware Support Application name per line.  
    The file should reside in the same folder as the Install-HSA.ps1 script.  
    The names can be found using the -LIST parameter.  

    -OFFLINE -FILE '<FileName>.txt'

.PARAMETER NAME

    Used to install one Hardware Support Application.  The name can be found 
    using the -LIST parameter.

    -OFFLINE -NAME '<HSA Name>'

.PARAMETER NOSMSTS

    Used when running in WinPE, but not in a Microsoft.SMS.TSEnvironment, 
    provided by Microsoft Endpoint Manager (MEM), System Center 
    Configuration Manager (SCCM), or Microsoft Deployment Toolkit (MDT).

    Call this parameter with the drive letter where the Windows partition has 
    the Windows folder already installed.

    -NOSMSTS '<Drive Letter>:'

.PARAMETER DEBUGINFORMATION

    Use to turn on Transcript logging and full logging of DISM Commands.  The 
    transcript file can be found at C:\Windows\Logs.  There will be a separate 
    log file from each DISM command generated at C:\Windows\Logs\DISM.

    -DEBUGINFORMATION

.EXAMPLE

    .\Install-HSA.ps1 -LIST

.EXAMPLE

    .\Install-HSA.ps1 -LIST -EXPORT

.EXAMPLE

    .\Install-HSA.ps1 -OFFLINE -NAME 'Lenovo Pen Settings'

.EXAMPLE

    .\Install-HSA.ps1 -OFFLINE -FILE 'List.txt'

.EXAMPLE

    .\Install-HSA.ps1 -OFFLINE -NOSMSTS 'D:' -ALL

.EXAMPLE

    .\Install-HSA.ps1 -OFFLINE -NOSMSTS 'D:' -ALL -DEBUGINFORMATION

.NOTES

    Return Code 1 = Both the -LIST and -OFFLINE commands were used.  Only one 
                    of these two parameters can be used at a time.
    Return Code 2 = More than one -ALL, -FILE, or -NAME were used.  Only one 
                    of these three parameters can be used at a time.
    Return Code 3 = No *_HSA_Manifest.json files were found in the 
                    subdirectories.
    Return Code 4 = When using the -FILE parameter, the file name was not 
                    found in the directory where the script resides.

#>

#######################
#  SCRIPT PARAMETERS  #
#######################
[CmdletBinding(DefaultParameterSetName = 'GetList')]
Param(
    [Parameter(ParameterSetName = 'GetList')]
    [switch]$List,
    [Parameter(ParameterSetName = 'GetList')]
    [switch]$Export,
    [Parameter(ParameterSetName = 'InstallOffline')]
    [switch]$Offline,
    [Parameter(ParameterSetName = 'InstallOffline')]
    [ValidateNotNullOrEmpty()]
    [string]$Name,
    [Parameter(ParameterSetName = 'InstallOffline')]
    [ValidateNotNullOrEmpty()]
    [string]$File,
    [Parameter(ParameterSetName = 'InstallOffline')]
    [switch]$All,
    [Parameter(ParameterSetName = 'InstallOffline')]
    [ValidateNotNullOrEmpty()]
    [string]$NoSMSTS,
    [Parameter(ParameterSetName = 'GetList')]
    [Parameter(ParameterSetName = 'InstallOffline')]
    [switch]$DebugInformation
)
###############
#  FUNCTIONS  #
###############
#Install-HSA
Function Install-HSA
{
    [CmdletBinding()]
    param (
        [PSCustomObject]$HSAPackage,
        [String]$HSAName
    )
    $OutDep = $Null
    If((($HSAName -contains $HSAPackage.hsa) -and ($Null -ne $HSAName))-or $All)
    {
        ForEach($Dep in $HSAPackage.Dependencies)
        {
            $OutDep += " /DependencyPackagePath:`"$($HSAPackage.JSONPath)\$($Dep)`""
        }
        $DISMLog = ""
        If($DebugInformation)
        {
            If(!(Test-Path -Path "$($LogPath)\DISM"))
            {
                New-Item -Path "$($LogPath)\DISM" -ItemType Directory
            }
            $DISMLog = " /LogLevel:4 /LogPath:`"$LogPath\DISM\$($HSAPackage.hsa).log`""
        }
        $DISMArgs = "/Add-ProvisionedAppxPackage /PackagePath:`"$($HSAPackage.JSONPath)\$($HSAPackage.appx)`" /LicensePath:`"$($HSAPackage.JSONPath)\$($HSAPackage.license)`"$($OutDep) /Region:`"All`"$DISMLog"
        Write-Host "Offline DISM - $($HSAPackage.hsa)"
        If($NoSMSTSPresent)
        {
            Write-Host "Using string data from NoSMSTS parameter to define the root drive letter for the DISM /Image parameter."
        }
        $DISMArgs = "/Image:$($Drive)\ $($DISMArgs)"
        Write-Host "$env:windir\system32\Dism.exe $DISMArgs"
        Start-Process -FilePath "$env:windir\system32\Dism.exe" -ArgumentList $DISMArgs -Wait
    }
}
##################
#  SCRIPT SETUP  #
##################
$FilePresent = $false
$NamePresent = $false
$NoSMSTSPresent = $false
If(($PSBoundParameters.ContainsKey('List') -and $PSBoundParameters.ContainsKey('Offline')))
{
    Write-Host "Use just one from the following list of parameters: -List or -Offline.  Review the script usage information for using these parameters."
    Return 1
}
If($Offline)
{
    If(($PSBoundParameters.ContainsKey('All') -and $PSBoundParameters.ContainsKey('File') -and $PSBoundParameters.ContainsKey('Name')) -or ($PSBoundParameters.ContainsKey('All') -and $PSBoundParameters.ContainsKey('File')) -or ($PSBoundParameters.ContainsKey('All') -and $PSBoundParameters.ContainsKey('Name')) -or ($PSBoundParameters.ContainsKey('File') -and $PSBoundParameters.ContainsKey('Name')) -or ((!($PSBoundParameters.ContainsKey('All')) -and (!($PSBoundParameters.ContainsKey('File'))) -and (!($PSBoundParameters.ContainsKey('Name'))))))
    {
        Write-Host "Use just one from the following list of parameters: -All, -Name, or -File.  Review the script usage information for using these parameters."
        Return 2
    }
    ElseIf($PSBoundParameters.ContainsKey('File'))
    {
        $FilePresent = $true
    }
    ElseIf($PSBoundParameters.ContainsKey('Name'))
    {
        $NamePresent = $true
    }
}
If($PSBoundParameters.ContainsKey('NoSMSTS'))
{
    $NoSMSTSPresent = $true
}
#Setup Vars
$ScriptDir = Split-Path $Script:MyInvocation.MyCommand.Path
If(($Offline) -and (!($NoSMSTSPresent)))
{
    $TSenv = New-Object -COMObject Microsoft.SMS.TSEnvironment
    If($TSEnv.value("OSDTargetSystemDrive") -ne "")
    {
        $Drive = "$($TSEnv.value("OSDTargetSystemDrive"))"
    }
    Else
    {
        $Drive = "$($TSEnv.value("OSDisk"))"
    }
}
ElseIf(($Offline) -and ($NoSMSTSPresent))
{
    $Drive = $NoSMSTS
}
ElseIf($List)
{
    $Drive = $env:SystemDrive
}
$LogDate = Get-Date -Format yyyyMMddHHmmss
If($DebugInformation)
{
    $LogPath = "$($Drive)\Windows\Logs"
    $LogFile = "$LogPath\$($myInvocation.MyCommand)_$($LogDate).log"

######################
#  START TRANSCRIPT  #
######################
    Start-Transcript $LogFile -Append -NoClobber
    Write-Host "Debug enabled"
}
##########
#  MAIN  #
##########
$MFJs = Get-ChildItem -path $ScriptDir -Recurse -File -Include "*_manifest.json"
If($Null -eq $MFJs)
{
    Write-Host "No HSA_Manifest.JSON files found in the subfolder structure."
    Return 3
}
Else
{
    $HSAPackages = @()
    ForEach($MFJ in $MFJs)
    {
        $MFJData = Get-Content -Path "$($MFJ.FullName)" | ConvertFrom-Json
        $HSAPackages += New-Object PSObject -Property @{'JSONPath' = $MFJ.DirectoryName; 'HSA' = $MFJData.HSA; 'Appx' = $MFJData.Appx; 'License' = $MFJData.License; 'Dependencies' = $MFJData.Dependencies}
    }
}
If($List -or $PSBoundParameters.Count -eq 0 -or ($DebugInformation -and $PSBoundParameters.Count -eq 1))
{
    If($PSBoundParameters.ContainsKey('Export'))
    {
        ForEach($Package in $HSAPackages)
        {
            $Package.hsa | Out-File "$ScriptDir\Export_$LogDate.txt" -Append -Noclobber
        }
    }
    Else
    {
        ForEach($Package in $HSAPackages)
        {
            Write-Host $Package.hsa
        }
    }
}
If($Offline)
{
    If($All)
    {
        Write-Host "Installing all HSAs found in the folder structure."
    }
    ElseIf($NamePresent)
    {
        Write-Host "Installing the $Name HSA."
    }
    ElseIf($FilePresent)
    {
        Write-Host "Reading the list of HSAs from $File."
        If(!(Test-Path -Path "$ScriptDir\$File"))
        {
            Write-Host "File: $File not found in $ScriptDir"
            Return 4
        }
    }
    ForEach($Package in $HSAPackages)
    {
        If($All)
        {
            Install-HSA -HSAPackage $Package
        }
        ElseIf($FilePresent -or $NamePresent)
        { 
            If($FilePresent)
            {
                $InstallFileArray = @()
                $InstallFileArray = Get-Content -Path "$ScriptDir\$File"
                ForEach($InstallFile in $InstallFileArray)
                {
                    Install-HSA -HSAPackage $Package -HSAName $InstallFile
                }
            }
            If($NamePresent)
            {
                Install-HSA -HSAPackage $Package -HSAName $Name
            }
        }
        $InstallFileArray = $Null
    }
}
If($DebugInformation)
{
    #####################
    #  STOP TRANSCRIPT  #
    #####################
    Stop-Transcript
}
