<#
 .SYNOPSIS
    generates an USB Stick to be used for a vanilla installation of Windows 10 on a new machine

 .DESCRIPTION
    generates an USB Stick to be used for a vanilla installation of Windows 10 on a new machine
	will copy custom files provided to further customize the installation from $OEM$ installation media
    Parameters all have defaults

 .PARAMETER SourcePath2ISO
    has to be the full path to the w10 installation ISO

 .PARAMETER DirWPE64
    has to be the full path to the WinPE AMD64 Environment Root Directory which will be created by this script using
	Create_WinPE64_Root.cmd, which in turn has to reside in the same directory as this script

 .PARAMETER DirCST_Files
    has to be the full path to the additional files to be copied to the sources\$OEM$ dir of the (NTFS) install media partition

 .PARAMETER DriverFiles2Add
    has to be the full path to the additional drivers to be added to the boot.wim on the FAT32 boot partition
    this should be WinPE drivers, if available and neccessary, e.g. USB 3.0 or NVMe storage drivers

 .PARAMETER USBSize
    has to be a long integer (e.g. [long]28gb, defaults to 28 gb

 .EXAMPLES
    .\prepare-usb-w10-install-1.ps1 -DirWPE64 "C:\temp\wpe64" -DirCST_Files "C:\temp\autounattends"
    this will use given strings to overwrite the defaults for the WinPE Root and for the custom files to be added

    .\prepare-usb-w10-install-1.ps1 -SourcePath2ISO "C:\temp\w10202h2de.iso"
    this will use the given filename as source for the w10 installation ISO

 .NOTES
    Permission level gets checked. It is assumed you start this script as Admin.
    An actual Windows 10 ADK has to be present (for the WinPE Build part) 
    The disk formatting and partitioning part will FORMAT the USB drive (WITH Confirmation)
    and will set the two drive letters to S (FAT32 Boot Partition) and Z (NTFS Install Media partition).
    Therefore it is STRONGLY recommended to have only ONE USB drive plugged in!

    requires PowerShell 4.0 upwards due to admin rights check

    Version 1.1 28.11.2020
    Copyright Mustafa Demirkiran & Rainer Pollmann
	mustafa@deploymentguru.de, rainer@deploymentguru.de
	Blog in https://deploymentguru.de

    Version 1.2 06.01.2021
    Bugfixing and Updates 

    Version 1.3 10.01.2021
    more requirement checks and verbose output to console 

    Version 1.4 18.02.2021
    updated / changed robocopy file jobs
    functionality to add drivers to boot.wim

    Version 1.5 27.02.2021
    disconnect network drives Z: and S: if present

    Version 1.6 30.04.2021
    add parameter for min size of USB media


#>


param(
     [Parameter()]
     [string]$SourcePath2ISO = "D:\temp\de_windows_10_business_editions_version_20h2_updated_feb_2021_x64_dvd_f32d90fa.iso",
 
     [Parameter()]
     [string]$DirWPE64 = "C:\TEMP\tools\WPE64",
     
     [Parameter()]
     [string]$DirCST_Files = "C:\TEMP\Autounattends",
     
     [Parameter()]
     [string]$DriverFiles2Add = "C:\DeploymentShare\Out-of-Box Drivers",

     [Parameter()]
     [long]$USBSize = 28gb

 )

#requires -RunAsAdministrator

$ErrorActionPreference = "SilentlyContinue"

# set defaults to be checked
$DriverFiles2Add = "C:\DeploymentShare\Out-of-Box Drivers"
$WimMountPath = $ENV:TEMP +"\wimmount"
$Features2AddPath = "C:\Program Files (x86)\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\amd64\WinPE_OCs"

$USBWinPE = "S:\sources\boot.wim"
$InstallWimFile = "Z:\sources\install.wim"
$ActionOK = $True
$W10ADKScript = "C:\Program Files (x86)\Windows Kits\10\Assessment and Deployment Kit\Deployment Tools\DandISetEnv.bat"
$WinPEADKScript = "C:\Program Files (x86)\Windows Kits\10\Assessment and Deployment Kit\Windows Preinstallation Environment\copype.cmd"

# default values may have been overwritten if corresponding (named) parameters were entered
$scriptPath = split-path -parent $MyInvocation.MyCommand.Definition
$Create_WinPE_Env = $scriptpath +"\Create_WinPE64_Root.cmd " + "amd64 " + $DirWPE64
$WinPEBuildScript = $scriptpath +"\Create_WinPE64_Root.cmd"

#remove network drives with Z and/or S if present
$nwdrives = Get-WmiObject -Class Win32_logicaldisk -Filter "DriveType = '4'"
$DrivesCount = $nwdrives.DeviceID.count
If ($DrivesCount -gt 0)
{
    $Index = 0
    Do
    {
        $DriveLetter = [String]$nwdrives.DeviceID[$Index]
        If (!($DriveLetter.contains(":")))
        {
        $DriveLetter = $DriveLetter + ":"
        }
        If (($Driveletter -eq "Z:") -or ($Driveletter -eq "S:"))
        {
            Write-Host "Drive found: " $DriveLetter ", will remove it now" -ForegroundColor Yellow
            $cmd = "net use " + $DriveLetter + " /delete"
            Invoke-Expression $cmd
        }
        $Index++
    }
    while ($Index -lt $DrivesCount)
}

# continue only if ADK and WinPE for ADK and Build Script are present
If ((Test-Path -Path $W10ADKScript) -and $ActionOK)
{
    $ActionOK = $True
    Write-Host "Windows 10 ADK found, continuing .." -ForeGroundColor green
}
Else
{
    $ActionOK = $False
    Write-Host "Windows 10 ADK ($W10ADKScript) NOT found, please install Windows 10 ADK first!" -ForeGroundColor red
    # exit
    Exit
}
If ((Test-Path -Path $WinPEADKScript) -and $ActionOK)
{
    $ActionOK = $True
    Write-Host "WinPE ADK AddOn found, continuing .." -ForeGroundColor green
}
Else
{
    $ActionOK = $False
    Write-Host "WinPE ADK Script $WinPEADKScript NOT found, please install Windows Preinstallation Environment for 10 ADK first!" -ForeGroundColor red
    # exit
    Exit
}
If ((Test-Path -Path $WinPEBuildScript) -and $ActionOK)
{
    $ActionOK = $True
    Write-Host "WinPE Build Script found, continuing .." -ForeGroundColor green
}
Else
{
    $ActionOK = $False
    Write-Host "WinPE Build Script $WinPEBuildScript NOT found, please make sure that it resides in the same directory as this script!" -ForeGroundColor red
    # exit
    Exit
}
If ((Test-Path -Path $SourcePath2ISO) -and $ActionOK)
{
    $ActionOK = $True
    Write-Host "Source ISO found, continuing .." -ForeGroundColor green
}
Else
{
    $ActionOK = $False
    Write-Host "Source ISO $SourcePath2ISO NOT found!" -ForeGroundColor red
    # exit
    Exit
}
If ($ActionOK)
{
    # check if an appropriate USB stick is attached
    $Disk = Get-Disk | Where-Object {$_.Path -match "USBSTOR" -and $_.Size -gt $USBSize -and -not $_.IsBoot }
    If ([string]::IsNullOrEmpty($Disk))
    {
        $ActionOK = $False
        Write-Host "no appropriate USB media attached, please attach an appropriate USB media with min size of $USBSize in Byte which is not bootable already!" -ForeGroundColor red
        # exit
        Exit
    }
    Else
    {
        $ActionOK = $True
        Write-Host "appropriate USB media found, continuing .." -ForeGroundColor green
    }
}

If ($ActionOK)
{
    # generate wipe 64 boot code source dir using ADK 10 standard scripts
    # Check if WinPE_Env exist already
	$WPEBootFile = $DirWPE64 + "\media\sources\boot.wim"
    If (Test-Path -path $WPEBootFile)
    {
        Write-Host "WinPE Environment already existing in Directory "  $DirWPE64 -ForegroundColor green
        $ActionOK = $True
    }
    Else
    {
        # Generate WPE Environment including new Path and variables
        Write-Host "Creating WinPE Environment in Directory  $DirWPE64 using $Create_WinPE_Env" -ForegroundColor yellow
        Invoke-Expression -Command $Create_WinPE_Env
    }

    # check correct USB drive for install media
    # if USB stick gets reused / plugged in to anotehr USB 3 port it may get the wrong drive letter (shuold be Z)
    # Check if WinPE_Env exist already
    If (Test-Path -path $InstallWimFile)
    {
        Write-Host "install.wim found on Install Media Partition (Z:\sources)" -ForegroundColor green
        $ActionOK = $True
    }
    Else
    {
        Write-Host "NO install.wim found on Install Media Partition (Z:\sources)!" -ForegroundColor magenta
        Write-Host "will try to create USB Media ..." -ForegroundColor yellow
        $ActionOK = $False
    }

        # Prepare USB Stick (only if not prepared already and install.wim was found!)
        If ((Test-Path -path $USBWinPE) -and ($ActionOK) )
            {
                Write-Host "WinPE boot.wim ($USBWinPE) is already existing on USB Media, will NOT prepare USB Media ..." -ForegroundColor green
            }
            Else
            {
                Write-Host "preparing USB Media now, WILL format this USB Media!" -ForegroundColor magenta
                # The following command will set $Disk to an USB drive with >28 GB of storage
			    # it is STRONGLY recommended to have only ONE USB drive plugged in (which will be partitioned and formated)!
                $Disk = Get-Disk | Where-Object {$_.Path -match "USBSTOR" -and $_.Size -gt 28Gb -and -not $_.IsBoot }
            
                # Clear the disk. This will delete any data on the disk!
			    # Validate that this is the correct disk that you want to completely erase!
                # To skip the confirmation prompt, set -confirm:$False
                Clear-Disk -InputObject $Disk[0] -RemoveData -confirm:$True
             
                # This command creates a 2047 MB system boot partition
                $SystemPartition = New-Partition -InputObject $Disk[0] -Size (2047MB) -IsActive
            
                # This formats the volume with a FAT32 Filesystem
                # To skip the confirmation dialog, set -Confirm:$False
                Format-Volume -NewFileSystemLabel "UFD-System" -FileSystem FAT32 -Partition $SystemPartition -Confirm:$True
            
                # This command creates an NTFS volume using the maximum space available on the drive.
                $OSPartition = New-Partition -InputObject $Disk[0] -UseMaximumSize
                Format-Volume -NewFileSystemLabel "UFD-Windows" -FileSystem NTFS -Partition $OSPartition -Confirm:$True
            
                # This command assigns drive letters to the new drive, the drive letters chosen should not already be in use.
                Set-Partition -InputObject $SystemPartition -NewDriveLetter "S"
                Set-Partition -InputObject $OSPartition -NewDriveLetter "Z"
            
            }

    # if formatting was NOT confirmed, do NOT continue
    If (Test-Path -path "S:\")
    {
        Write-Host "USB Boot FAT32 Partition was created successfully on Drive S: or is already present." -ForegroundColor green
        $ActionOK= $True
    }
    Else
    {
        Write-Host "USB Media creation was CANCELLED, no further actions!" -ForegroundColor red
        $ActionOK = $False
    }
    If (Test-Path -path "Z:\")
    {
        Write-Host "NTFS Installation Media Partition was created successfully on Drive Z: or is already present." -ForegroundColor green
        $ActionOK= $True
    }
    Else
    {
        Write-Host "USB Media creation was CANCELLED, no further actions!" -ForegroundColor red
        $ActionOK = $False
    }


    If ($ActionOK)
    {
        # Generate Source Directory Content
        # mount OS ISO
        Write-Host "mounting SourcePath Win10 source $SourcePath2ISO ..." -ForegroundColor yellow
        $mounting = Mount-DiskImage -ImagePath $SourcePath2ISO
        $Drive = (Get-DiskImage -ImagePath $SourcePath2ISO | Get-Volume).DriveLetter
        # check if an install.wim is part of the mounted ISO
        $PathtoSourceWIM = $Drive + ":\sources\install.wim"
        If (Test-Path -path $PathtoSourceWIM)
        {
            $ActionOK = $True
            Write-Host "Install.WIM in Source $SourcePath2ISO found ..." -ForeGroundColor green
        }
        Else
        {
            $ActionOK = $False
            Write-Host "NO Install.WIM was found in Source $SourcePath2ISO ..." -ForeGroundColor red
            # Unmount ISO
            Write-Host "unmounting Win10 Source $SourcePath2ISO ,,," -ForegroundColor yellow
            Dismount-DiskImage -ImagePath $SourcePath2ISO
        }


        # continue with copying only, if all requirements are met
        If ($ActionOK)
        {
            # Copy / Update WPE Sources to USB
            # prepare robocopy commands
            $CMD1 = "robocopy.exe " + $Drive + ":\ " + "Z:\ *.* /S"
            $CMD2 = "robocopy.exe " + $DirWPE64 + "\media " + "s:\ *.* /S /XF boot.wim"
            $CMD3 = "robocopy.exe " + $Drive + ":\sources " + "s:\sources boot.wim"
            $CMD4 = "robocopy.exe " + $Drive + ":\ " + "s:\ setup.exe" 
            $CMD5 = "robocopy.exe " + $DirCST_Files + " 'Z:\sources\`$OEM$' /S /XF *unattend*.*"
            $CMD6 = "robocopy.exe " + $DirCST_Files + " S:\  autounattend.xml"
            $CMD7 = "robocopy.exe " + $DirCST_Files + " Z:\  autounattend.xml"

            Write-Host "copying files from source ISO and from WinPE root, will take some time ..." -ForegroundColor yellow
            # Call Robocopy CMDs (will take some time)
            # copy all sources from mounted ISO to NTFS partition except boot.wim and setup.exe
            Write-Host "copy all sources from mounted ISO to NTFS partition ..." -ForegroundColor yellow
            Invoke-Expression $CMD1
            # copy all files from WinPE Media Root to boot drive root except winpe boot.wim
            Write-Host "copy all files from WinPE Media Root to boot drive root except winpe boot.wim ..." -ForegroundColor yellow
            # Write-Host "copy all files from WinPE Media Root to boot drive root ..." -ForegroundColor yellow
            Invoke-Expression $CMD2
            # copy original boot.wim from mounted setup iso to sources dir boot drive
            # this will use the windows 10 boot image INSTEAD of the universal winpe boot image
            Write-Host "copy original boot.wim from mounted setup iso to sources dir to boot drive ..." -ForegroundColor yellow
            Invoke-Expression $CMD3
            # copy setup.exe from mounted installation media (ISO) to boot drive root
            Write-Host "copy setup.exe from mounted installation media (ISO) to boot drive root ..." -ForegroundColor yellow
            Invoke-Expression $CMD4

            # Copy / Update Custom Specific Files
            If (!(Test-Path -path $DirCST_Files))
                {
                    Write-Host "No Custom Files found in $DirCST_Files!" -ForegroundColor magenta
                }
            Else
            {
                # copy customization files to installation media NTFS partition into ./sources/$OEM$ EXCEPT autounattend scripts
                Write-Host "copying customization files ..." -ForegroundColor yellow
                Invoke-Expression $CMD5
                # copy autounattend script from custom files root to boot drive root
                Write-Host "copy autounattend script from custom files root to boot drive root ..." -ForegroundColor yellow
                Invoke-Expression $CMD6
                # copy autounattend script from custom files root to NTFS USB Drive root too 
                Write-Host "copy autounattend script from custom files root to NTFS USB Drive root ..." -ForegroundColor yellow
                Invoke-Expression $CMD7
            }
            # Unmount ISO
            Write-Host "unmounting Win10 Source $SourcePath2ISO ..." -ForegroundColor yellow
            Dismount-DiskImage -ImagePath $SourcePath2ISO

            # add drivers to boot.wim in fat32 partition
            If(Test-Path $DriverFiles2Add)
            {
                Write-Host "adding drivers to boot.wim $USBWinPE from $DriverFiles2Add ..." -ForegroundColor yellow
                # change readonly flag of boot wim
                $file = Get-Item -Path $USBWinPE
                $file.IsReadOnly = $false
                # make mount dir
				# remove mount dir (for potentially next run with same user context) before, to be sure
                If (Test-Path $WimMountPath)
				    {
                        Remove-Item -LiteralPath $WimMountPath -Force -Recurse
					}
                New-Item -Path $WimMountPath -ItemType Directory
                # mount boot wim
                Mount-WindowsImage -ImagePath $USBWinPE -Index 1 -Path $WimMountPath
                # add drivers
                Add-WindowsDriver -Path $WimMountPath -Driver $DriverFiles2Add -Recurse -ForceUnsigned
                # save changes made to wim and unmount wim
                Dismount-WindowsImage -Path $WimMountPath -Save
            }
            Write-Host "done, you can eject the USB media now!" -ForegroundColor green
        }
    }
}

# SIG # Begin signature block
# MIIF3AYJKoZIhvcNAQcCoIIFzTCCBckCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUk8k+Gc7thOZNSVpsTjGXjfOt
# lZWgggNSMIIDTjCCAjagAwIBAgIQO6xCYsZPLptB4/45+FXF+zANBgkqhkiG9w0B
# AQsFADA/MQswCQYDVQQGEwJERTEYMBYGA1UECgwPQ0MgRGlzdHJpYnV0aW9uMRYw
# FAYDVQQDDA1Db21wdXRhY2VudGVyMB4XDTIxMDEyNjEwMTc0NloXDTI1MDEyNjEw
# Mjc0NlowPzELMAkGA1UEBhMCREUxGDAWBgNVBAoMD0NDIERpc3RyaWJ1dGlvbjEW
# MBQGA1UEAwwNQ29tcHV0YWNlbnRlcjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCC
# AQoCggEBAMLt6nLy4XMYSeyJkTAOV8OpipL1EYW/F8uqp8IBQIs04SpdhodYZLPe
# tFoJmPPP+Iv1hU5x3dH2CSKM4Y36E1PviYGrXzu78Qqa5BEQpqAEw3nsRFJd4SNd
# 0PwjWP9Ns31S22Uf3yH9hthRuDY/nAnCoyl4T4S9VvqobaUNk7s4aCupoOcorgQw
# BAAvdgE3ucCdjohvBrs4zTBKqvsSJjHgll7GqmVdoHdiZZrWwFDCm7cswP7KMWcB
# LIKmjau2F8EqFo1ZuGKPDxB/CHywAJPPRksuywT1foEY5l2zngqTIbBOarGZBpLq
# 68ABM6tMIfuFJFMVP9oljMxosKDy+NECAwEAAaNGMEQwDgYDVR0PAQH/BAQDAgeA
# MBMGA1UdJQQMMAoGCCsGAQUFBwMDMB0GA1UdDgQWBBRlY+0Y8olATZspek7tOp27
# geNclDANBgkqhkiG9w0BAQsFAAOCAQEAbLN/Qr1r+rzzeqpzQTDaI8AIPJbpa3GL
# RJkdsRQRLbkKDhBbjyu+Z6iYciHF2iYbnLH9g5VnF3FotPO+cpYoCYZt7RMgoptA
# XkZEgKcnvWU39ag5aFfC0bUAo4AakAp0CQiIUx8ioxKWmWxIKLpZiJSWqjPFGLKv
# CxQy3+2mfnOzvhvBC2FZwkODMgD5KhImi0LnUU7FZ212w99wfbVszQsqYI9JVsib
# M12nbp0EWDYv6jIiW8zIxiBUi9poqeLm8rgvWKGJ5L7Yb5mVXyVeNhKo+cbk5ysm
# 1DtTvqcOuBnacUVItbkUb/7lCRZVXJoE2IJ6KZwoxlwIi5yYYiMOszGCAfQwggHw
# AgEBMFMwPzELMAkGA1UEBhMCREUxGDAWBgNVBAoMD0NDIERpc3RyaWJ1dGlvbjEW
# MBQGA1UEAwwNQ29tcHV0YWNlbnRlcgIQO6xCYsZPLptB4/45+FXF+zAJBgUrDgMC
# GgUAoHgwGAYKKwYBBAGCNwIBDDEKMAigAoAAoQKAADAZBgkqhkiG9w0BCQMxDAYK
# KwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4wDAYKKwYBBAGCNwIBFTAjBgkqhkiG
# 9w0BCQQxFgQU3QTl2O3wVep9mh5rDdn67pjGxdcwDQYJKoZIhvcNAQEBBQAEggEA
# oZXkFfLxgKpZ0qVVmPt+4pFSIWHvfsU9vjCvzf6AC89bqb3WRj62NcSU1b16igCL
# SvNMkuRPIjBHaT684H9Q8SW0m6JBA78TNVVtE40Rm1DD56utJOKLQb8cAO4UCLrs
# G2vxi+tgvJm/pkBQZzg2F25EcOgOGGZhRp+jOjAeBrOucAFE1hRKZiDZYBtK3bWa
# bJwkVowTIi7XvV5w52zuqHnVHQQFOCcTpgAo9v9ilAN7fJNJwTnJDMB1ELK//XKF
# zRzgx3MpvhcIe+oFWlV/QY8iiG8HwF7X98n0cI/oX8VkcqOkc0grYPhAqmjhpDic
# x7AHSZUKlT+Uu0JMtdEeSg==
# SIG # End signature block
